"""
Settings used by django-python3.
"""
from ssl import PROTOCOL_TLS

from django.conf import settings


class LazySetting(object):

    """
    A proxy to a named Django setting.
    """

    def __init__(self, name, default=None):
        self.name = name
        self.default = default

    def __get__(self, obj, cls):
        if obj is None:
            return self
        return getattr(obj._settings, self.name, self.default)


class LazySettings(object):

    """
    A proxy to ldap-specific django settings.

    Settings are resolved at runtime, allowing tests
    to change settings at runtime.
    """

    def __init__(self, settings):
        self._settings = settings

    LDAP_AUTH_URL = LazySetting(
        name="LDAP_AUTH_URL",
        default=["ldap://localhost:389"],
    )

    LDAP_AUTH_USE_TLS = LazySetting(
        name="LDAP_AUTH_USE_TLS",
        default=False,
    )

    LDAP_AUTH_TLS_CIPHERS = LazySetting(
        name="LDAP_AUTH_TLS_CIPHERS",
        default="ALL",
    )

    LDAP_AUTH_TLS_VERSION = LazySetting(
        name="LDAP_AUTH_TLS_VERSION",
        default=PROTOCOL_TLS,
    )

    LDAP_AUTH_TLS_ARGS = LazySetting(
        name="LDAP_AUTH_TLS_ARGS",
        default={},
    )

    LDAP_AUTH_SEARCH_BASE = LazySetting(
        name="LDAP_AUTH_SEARCH_BASE",
        default="ou=people,dc=example,dc=com",
    )

    LDAP_AUTH_OBJECT_CLASS = LazySetting(
        name="LDAP_AUTH_OBJECT_CLASS",
        default="inetOrgPerson",
    )

    LDAP_AUTH_USER_FIELDS = LazySetting(
        name="LDAP_AUTH_USER_FIELDS",
        default={
            "username": "uid",
            "first_name": "givenName",
            "last_name": "sn",
            "email": "mail",
        },
    )

    LDAP_AUTH_USER_LOOKUP_FIELDS = LazySetting(
        name="LDAP_AUTH_USER_LOOKUP_FIELDS",
        default=(
            "username",
        ),
    )

    LDAP_AUTH_CLEAN_USER_DATA = LazySetting(
        name="LDAP_AUTH_CLEAN_USER_DATA",
        default="django_python3_ldap.utils.clean_user_data",
    )

    LDAP_AUTH_FORMAT_SEARCH_FILTERS = LazySetting(
        name="LDAP_AUTH_FORMAT_SEARCH_FILTERS",
        default="django_python3_ldap.utils.format_search_filters",
    )

    LDAP_AUTH_SYNC_USER_RELATIONS = LazySetting(
        name="LDAP_AUTH_SYNC_USER_RELATIONS",
        default="django_python3_ldap.utils.sync_user_relations",
    )

    LDAP_AUTH_FORMAT_USERNAME = LazySetting(
        name="LDAP_AUTH_FORMAT_USERNAME",
        default="django_python3_ldap.utils.format_username_openldap",
    )

    LDAP_AUTH_ACTIVE_DIRECTORY_DOMAIN = LazySetting(
        name="LDAP_AUTH_ACTIVE_DIRECTORY_DOMAIN",
        default=None,
    )

    LDAP_AUTH_TEST_USER_USERNAME = LazySetting(
        name="LDAP_AUTH_TEST_USER_USERNAME",
        default="",
    )

    LDAP_AUTH_TEST_USER_EMAIL = LazySetting(
        name="LDAP_AUTH_TEST_USER_EMAIL",
        default=""
    )

    LDAP_AUTH_TEST_USER_PASSWORD = LazySetting(
        name="LDAP_AUTH_TEST_USER_PASSWORD",
        default="",
    )

    LDAP_AUTH_CONNECTION_USERNAME = LazySetting(
        name="LDAP_AUTH_CONNECTION_USERNAME",
        default=None,
    )

    LDAP_AUTH_CONNECTION_PASSWORD = LazySetting(
        name="LDAP_AUTH_CONNECTION_PASSWORD",
        default=None,
    )

    LDAP_AUTH_CONNECT_ARGS = LazySetting(
        name="LDAP_AUTH_CONNECT_ARGS",
        default={},
    )

    LDAP_AUTH_CONNECT_USE_SSL = LazySetting(
        name="LDAP_AUTH_CONNECT_USE_SSL",
        default=False,
    )

    LDAP_AUTH_CONNECT_TIMEOUT = LazySetting(
        name="LDAP_AUTH_CONNECT_TIMEOUT",
        default=None
    )

    LDAP_AUTH_RECEIVE_TIMEOUT = LazySetting(
        name="LDAP_AUTH_RECEIVE_TIMEOUT",
        default=None
    )

    LDAP_AUTH_POOL_ACTIVE = LazySetting(
        name="LDAP_AUTH_POOL_ACTIVE",
        default=True
    )


settings = LazySettings(settings)



# encoding=utf-8
from __future__ import unicode_literals

from unittest import skipUnless, skip
from io import StringIO

from django.test import TestCase, override_settings
from django.contrib.auth import authenticate
from django.contrib.auth.models import User
from django.conf import settings as django_settings
from django.core.management import call_command, CommandError

from django_python3_ldap.conf import settings
from django_python3_ldap.ldap import connection
from django_python3_ldap.utils import clean_ldap_name, import_func


@skipUnless(settings.LDAP_AUTH_TEST_USER_USERNAME, "No settings.LDAP_AUTH_TEST_USER_USERNAME supplied.")
@skipUnless(settings.LDAP_AUTH_TEST_USER_PASSWORD, "No settings.LDAP_AUTH_TEST_USER_PASSWORD supplied.")
@skipUnless(settings.LDAP_AUTH_USER_LOOKUP_FIELDS == ("username",), "Cannot test using custom lookup fields.")
@skipUnless(django_settings.AUTH_USER_MODEL == "auth.User", "Cannot test using a custom user model.")
class TestLdap(TestCase):

    def setUp(self):
        super(TestLdap, self).setUp()
        User.objects.all().delete()

    # Lazy settings tests.

    def testLazySettingsInstanceLookup(self):
        self.assertTrue(settings.LDAP_AUTH_TEST_USER_USERNAME)

    def testLazySettingsClassLookup(self):
        self.assertEqual(settings.__class__.LDAP_AUTH_TEST_USER_USERNAME.name, "LDAP_AUTH_TEST_USER_USERNAME")
        self.assertEqual(settings.__class__.LDAP_AUTH_TEST_USER_USERNAME.default, "")

    # Utils tests.

    def testCleanLdapName(self):
        self.assertEqual(clean_ldap_name("foo@bar.com"), r'foo@bar.com')
        self.assertEqual(clean_ldap_name("café"), r'caf\E9')

    # LDAP tests.

    def testGetUserKwargsSuccess(self):
        with connection() as c:
            user = c.get_user(
                username=settings.LDAP_AUTH_TEST_USER_USERNAME,
            )
            self.assertIsInstance(user, User)
            self.assertEqual(user.username, settings.LDAP_AUTH_TEST_USER_USERNAME)

    def testGetUserKwargsIncorrectUsername(self):
        with connection() as c:
            user = c.get_user(
                username="bad" + settings.LDAP_AUTH_TEST_USER_USERNAME,
            )
            self.assertEqual(user, None)

    def testHasUserKwargsSuccess(self):
        with connection() as c:
            exist = c.has_user(
                username=settings.LDAP_AUTH_TEST_USER_USERNAME,
            )
            self.assertEqual(exist, True)

    def testHasUserKwargsIncorrectUsername(self):
        with connection() as c:
            exist = c.has_user(
                username="bad" + settings.LDAP_AUTH_TEST_USER_USERNAME,
            )
            self.assertEqual(exist, False)

    # Authentication tests.

    def testAuthenticateUserSuccess(self):
        user = authenticate(
            username=settings.LDAP_AUTH_TEST_USER_USERNAME,
            password=settings.LDAP_AUTH_TEST_USER_PASSWORD,
        )
        self.assertIsInstance(user, User)
        self.assertEqual(user.username, settings.LDAP_AUTH_TEST_USER_USERNAME)

    def testAuthenticateUserBadUsername(self):
        user = authenticate(
            username="bad" + settings.LDAP_AUTH_TEST_USER_USERNAME,
            password=settings.LDAP_AUTH_TEST_USER_PASSWORD,
        )
        self.assertEqual(user, None)

    def testAuthenticateUserBadPassword(self):
        user = authenticate(
            username=settings.LDAP_AUTH_TEST_USER_USERNAME,
            password="bad" + settings.LDAP_AUTH_TEST_USER_PASSWORD,
        )
        self.assertEqual(user, None)

    def testAuthenticateWithAdditonalKwargsUserSuccess(self):
        user = authenticate(
            username=settings.LDAP_AUTH_TEST_USER_USERNAME,
            password=settings.LDAP_AUTH_TEST_USER_PASSWORD,
            another_kwarg="whatever",
        )
        self.assertIsInstance(user, User)
        self.assertEqual(user.username, settings.LDAP_AUTH_TEST_USER_USERNAME)

    def testRepeatedUserAuthenticationDoestRecreateUsers(self):
        user_1 = authenticate(
            username=settings.LDAP_AUTH_TEST_USER_USERNAME,
            password=settings.LDAP_AUTH_TEST_USER_PASSWORD,
        )
        user_2 = authenticate(
            username=settings.LDAP_AUTH_TEST_USER_USERNAME,
            password=settings.LDAP_AUTH_TEST_USER_PASSWORD,
        )
        # Ensure that the user isn't recreated on second access.
        self.assertEqual(user_1.pk, user_2.pk)

    @skip("FIXME: test server currently uses outdated TLS cyphers")
    def testAuthenticateWithTLS(self):
        with self.settings(LDAP_AUTH_USE_TLS=True):
            user = authenticate(
                username=settings.LDAP_AUTH_TEST_USER_USERNAME,
                password=settings.LDAP_AUTH_TEST_USER_PASSWORD,
            )
            self.assertIsInstance(user, User)
            self.assertEqual(user.username, settings.LDAP_AUTH_TEST_USER_USERNAME)

    def testAuthenticateWithRebind(self):
        with self.settings(
            LDAP_AUTH_CONNECTION_USERNAME=settings.LDAP_AUTH_TEST_USER_USERNAME,
            LDAP_AUTH_CONNECTION_PASSWORD=settings.LDAP_AUTH_TEST_USER_PASSWORD,
        ):
            user = authenticate(
                username=settings.LDAP_AUTH_TEST_USER_USERNAME,
                password=settings.LDAP_AUTH_TEST_USER_PASSWORD,
            )
            self.assertIsInstance(user, User)
            self.assertEqual(user.username, settings.LDAP_AUTH_TEST_USER_USERNAME)

    def testAuthenticateWithFailedRebind(self):
        with self.settings(
            LDAP_AUTH_CONNECTION_USERNAME="bad" + settings.LDAP_AUTH_TEST_USER_USERNAME,
            LDAP_AUTH_CONNECTION_PASSWORD=settings.LDAP_AUTH_TEST_USER_PASSWORD,
        ):
            user = authenticate(
                username=settings.LDAP_AUTH_TEST_USER_USERNAME,
                password=settings.LDAP_AUTH_TEST_USER_PASSWORD,
            )
            self.assertIs(user, None)

    def testAuthenticateWithLimitedRetries(self):
        # simulate offline server
        with self.settings(
            LDAP_AUTH_URL=["ldap://example.com:389"],
            LDAP_AUTH_POOL_ACTIVE=1,
        ):
            user = authenticate(
                username=settings.LDAP_AUTH_TEST_USER_USERNAME,
                password=settings.LDAP_AUTH_TEST_USER_PASSWORD,
            )
        self.assertEqual(user, None)

    # User synchronisation.

    def testSyncUsersCreatesUsers(self):
        call_command("ldap_sync_users", verbosity=0)
        self.assertGreater(User.objects.count(), 0)

    def testSyncUserWithLookup(self):
        call_command("ldap_sync_users", settings.LDAP_AUTH_TEST_USER_USERNAME, verbosity=0)
        self.assertEqual(User.objects.filter(username=settings.LDAP_AUTH_TEST_USER_USERNAME).count(), 1)

    @override_settings(LDAP_AUTH_USER_LOOKUP_FIELDS=('username', 'email'))
    def testSyncUserWithMultipleLookups(self):
        call_command(
            "ldap_sync_users",
            settings.LDAP_AUTH_TEST_USER_USERNAME,
            settings.LDAP_AUTH_TEST_USER_EMAIL,
            verbosity=0
        )
        self.assertEqual(User.objects.filter(username=settings.LDAP_AUTH_TEST_USER_USERNAME).count(), 1)

    def testSyncUsersCommandOutput(self):
        out = StringIO()
        call_command("ldap_sync_users", verbosity=1, stdout=out)
        rows = out.getvalue().split("\n")[:-1]
        self.assertEqual(len(rows), User.objects.count())
        for row in rows:
            self.assertRegex(row, r'^Synced ')

    def testReSyncUsersDoesntRecreateUsers(self):
        call_command("ldap_sync_users", verbosity=0)
        user_count_1 = User.objects.count()
        call_command("ldap_sync_users", verbosity=0)
        user_count_2 = User.objects.count()
        self.assertEqual(user_count_1, user_count_2)

    # User promotion.

    def testPromoteUser(self):
        user = User.objects.create(
            username="test",
        )
        self.assertFalse(user.is_staff)
        self.assertFalse(user.is_superuser)
        # Promote the user.
        call_command("ldap_promote", "test", stdout=StringIO())
        user = User.objects.get(username="test")
        self.assertTrue(user.is_staff)
        self.assertTrue(user.is_superuser)

    def testPromoteMissingUser(self):
        with self.assertRaises(CommandError, msg="User with username missing_user does not exist"):
            call_command("ldap_promote", "missing_user", verbosity=0)

    def testSyncUserRelations(self):
        def check_sync_user_relation(user, data, *, connection=None, dn=None):
            # id have been created
            self.assertIsNotNone(user.id)
            # connection was passed
            self.assertIsNotNone(connection)
            # dn was passed
            self.assertIsNotNone(dn)
            # model is saved
            self.assertEqual(user.username, User.objects.get(pk=user.id).username)
            # save all groups
            self.assertIn('cn', data)
            ldap_groups = list(data.get('memberOf', ()))
            ldap_groups.append('default_group')
            for group in ldap_groups:
                user.groups.create(name=group)

        with self.settings(LDAP_AUTH_SYNC_USER_RELATIONS=check_sync_user_relation):
            user = authenticate(
                username=settings.LDAP_AUTH_TEST_USER_USERNAME,
                password=settings.LDAP_AUTH_TEST_USER_PASSWORD,
            )
            self.assertIsInstance(user, User)
            self.assertGreaterEqual(user.groups.count(), 1)
            self.assertEqual(user.groups.filter(name='default_group').count(), 1)

    def testOldSyncUserRelations(self):
        def check_sync_user_relation(user, data):
            # id have been created
            self.assertIsNotNone(user.id)
            # model is saved
            self.assertEqual(user.username, User.objects.get(pk=user.id).username)
            # save all groups
            self.assertIn('cn', data)
            ldap_groups = list(data.get('memberOf', ()))
            ldap_groups.append('default_group')
            for group in ldap_groups:
                user.groups.create(name=group)

        with self.settings(LDAP_AUTH_SYNC_USER_RELATIONS=check_sync_user_relation):
            user = authenticate(
                username=settings.LDAP_AUTH_TEST_USER_USERNAME,
                password=settings.LDAP_AUTH_TEST_USER_PASSWORD,
            )
            self.assertIsInstance(user, User)
            self.assertGreaterEqual(user.groups.count(), 1)
            self.assertEqual(user.groups.filter(name='default_group').count(), 1)

    def testImportFunc(self):
        self.assertIs(clean_ldap_name, import_func(clean_ldap_name))
        self.assertIs(clean_ldap_name, import_func('django_python3_ldap.utils.clean_ldap_name'))
        self.assertTrue(callable(import_func('django.contrib.auth.get_user_model')))

        self.assertRaises(AttributeError, import_func, 123)

        self.assertTrue(callable(import_func(settings.LDAP_AUTH_SYNC_USER_RELATIONS)))

        with self.settings(LDAP_AUTH_SYNC_USER_RELATIONS='django.contrib.auth.get_user_model'):
            self.assertTrue(callable(import_func(settings.LDAP_AUTH_SYNC_USER_RELATIONS)))

    def testCleanUsersDeactivate(self):
        """
        ldap_clean_users management command test
        """
        from django.contrib.auth import get_user_model
        User = get_user_model()
        _username = "nonldap{user}".format(user=settings.LDAP_AUTH_TEST_USER_USERNAME)
        user = User.objects.create_user(
            _username,
            "nonldap{mail}".format(mail=settings.LDAP_AUTH_TEST_USER_EMAIL),
            settings.LDAP_AUTH_TEST_USER_PASSWORD)
        user.save()
        user_count_1 = User.objects.count()
        self.assertEqual(User.objects.get(username=_username).is_active, True)
        call_command("ldap_clean_users", verbosity=0)
        user_count_2 = User.objects.count()
        self.assertEqual(user_count_1, user_count_2)
        self.assertEqual(User.objects.get(username=_username).is_active, False)

        """
        Test with lookup
        """
        # Reactivate user
        user = User.objects.get(username=_username)
        user.is_active = True
        user.save()
        # Create second user
        _usernameLookup = "nonldaplookup{user}".format(user=settings.LDAP_AUTH_TEST_USER_USERNAME)
        user = User.objects.create_user(
            _usernameLookup,
            "nonldaplookup{mail}".format(mail=settings.LDAP_AUTH_TEST_USER_EMAIL),
            settings.LDAP_AUTH_TEST_USER_PASSWORD)
        user.save()
        user_count_1 = User.objects.count()
        self.assertEqual(User.objects.get(username=_usernameLookup).is_active, True)
        # Clean second user
        call_command("ldap_clean_users", _usernameLookup, verbosity=0)
        user_count_2 = User.objects.count()
        self.assertEqual(user_count_1, user_count_2)
        self.assertEqual(User.objects.get(username=_usernameLookup).is_active, False)
        self.assertEqual(User.objects.get(username=_username).is_active, True)
        # Reactivate second user
        user = User.objects.get(username=_usernameLookup)
        user.is_active = True
        user.save()
        # Clean first user
        call_command("ldap_clean_users", _username, verbosity=0)
        self.assertEqual(User.objects.get(username=_username).is_active, False)
        self.assertEqual(User.objects.get(username=_usernameLookup).is_active, True)
        # Lookup a non existing user (raise a CommandError)
        with self.assertRaises(CommandError):
            call_command("ldap_clean_users", 'doesnonexist', verbosity=0)

        """
        Test with superuser
        """
        # Reactivate first user and promote to superuser
        user = User.objects.get(username=_username)
        user.is_active = True
        user.is_superuser = True
        user.save()
        # Reactivate second user
        user = User.objects.get(username=_usernameLookup)
        user.is_active = True
        user.save()
        call_command("ldap_clean_users", superuser=False, verbosity=0)
        self.assertEqual(User.objects.get(username=_username).is_active, True)
        self.assertEqual(User.objects.get(username=_usernameLookup).is_active, False)
        call_command("ldap_clean_users", superuser=True, verbosity=0)
        self.assertEqual(User.objects.get(username=_username).is_active, False)

        """
        Test with staff user
        """
        # Reactivate first user and promote to staff
        user = User.objects.get(username=_username)
        user.is_active = True
        user.is_superuser = False
        user.is_staff = True
        user.save()
        # Reactivate second user
        user = User.objects.get(username=_usernameLookup)
        user.is_active = True
        user.save()
        call_command("ldap_clean_users", staff=False, verbosity=0)
        self.assertEqual(User.objects.get(username=_username).is_active, True)
        self.assertEqual(User.objects.get(username=_usernameLookup).is_active, False)
        call_command("ldap_clean_users", staff=True, verbosity=0)
        self.assertEqual(User.objects.get(username=_username).is_active, False)

    def testCleanUsersPurge(self):
        """
        ldap_clean_users management command test with purge argument
        """
        from django.contrib.auth import get_user_model
        User = get_user_model()
        user = User.objects.create_user(
            "nonldap{user}".format(user=settings.LDAP_AUTH_TEST_USER_USERNAME),
            "nonldap{mail}".format(mail=settings.LDAP_AUTH_TEST_USER_EMAIL),
            settings.LDAP_AUTH_TEST_USER_PASSWORD)
        user.save()
        user_count_1 = User.objects.count()
        call_command("ldap_clean_users", verbosity=0, purge=True)
        user_count_2 = User.objects.count()
        self.assertGreater(user_count_1, user_count_2)

    def testCleanUsersCommandOutput(self):
        # Test without purge
        out = StringIO()
        from django.contrib.auth import get_user_model
        User = get_user_model()
        user = User.objects.create_user(
            "nonldap{user}".format(user=settings.LDAP_AUTH_TEST_USER_USERNAME),
            "nonldap{mail}".format(mail=settings.LDAP_AUTH_TEST_USER_EMAIL),
            settings.LDAP_AUTH_TEST_USER_PASSWORD)
        user.save()
        call_command("ldap_clean_users", stdout=out, verbosity=1)
        rows = out.getvalue().split("\n")[:-1]
        self.assertEqual(len(rows), 1)
        for row in rows:
            self.assertRegex(row, r'^Deactivated ')
        # Reset for next test
        user.delete()
        out.truncate(0)
        out.seek(0)
        # Test with purge
        user = User.objects.create_user(
            "nonldap{user}".format(user=settings.LDAP_AUTH_TEST_USER_USERNAME),
            "nonldap{mail}".format(mail=settings.LDAP_AUTH_TEST_USER_EMAIL),
            settings.LDAP_AUTH_TEST_USER_PASSWORD)
        user.save()
        call_command("ldap_clean_users", stdout=out, verbosity=1, purge=True)
        rows = out.getvalue().split("\n")[:-1]
        self.assertEqual(len(rows), 1)
        for row in rows:
            self.assertRegex(row, r'^Purged ')

    def testReCleanUsersDoesntRecreateUsers(self):
        from django.contrib.auth import get_user_model
        User = get_user_model()
        user = User.objects.create_user(
            "nonldap{user}".format(user=settings.LDAP_AUTH_TEST_USER_USERNAME),
            "nonldap{mail}".format(mail=settings.LDAP_AUTH_TEST_USER_EMAIL),
            settings.LDAP_AUTH_TEST_USER_PASSWORD)
        user.save()
        call_command("ldap_clean_users", verbosity=0, purge=True)
        user_count_1 = User.objects.count()
        call_command("ldap_clean_users", verbosity=0, purge=True)
        user_count_2 = User.objects.count()
        self.assertEqual(user_count_1, user_count_2)



from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand, CommandError
from django.db import transaction

from django_python3_ldap import ldap
from django_python3_ldap.conf import settings
from django_python3_ldap.utils import group_lookup_args


class Command(BaseCommand):

    help = "Creates local user models for users found in the remote LDAP authentication server."

    def add_arguments(self, parser):
        parser.add_argument(
            'lookups',
            nargs='*',
            type=str,
            help='A list of lookup values, matching the fields specified in LDAP_AUTH_USER_LOOKUP_FIELDS. '
                 'If this is not provided then ALL users are synced.'
        )

    @staticmethod
    def _iter_synced_users(connection, lookups):
        """
        Iterates over synced users. If the list of lookups is empty, then all users are synced using iter_users.
        However, if lookups are provided, get_user is used to sync each user found using the lookups.
        """
        if len(lookups) < 1:
            for user in connection.iter_users():
                yield user
        else:
            for lookup in group_lookup_args(*lookups):
                yield connection.get_user(**lookup)

    @transaction.atomic()
    def handle(self, *args, **kwargs):
        verbosity = int(kwargs.get("verbosity", 1))
        lookups = kwargs.get('lookups', [])
        User = get_user_model()
        auth_kwargs = {
            User.USERNAME_FIELD: settings.LDAP_AUTH_CONNECTION_USERNAME,
            'password': settings.LDAP_AUTH_CONNECTION_PASSWORD
        }
        with ldap.connection(**auth_kwargs) as connection:
            if connection is None:
                raise CommandError("Could not connect to LDAP server")
            for user in self._iter_synced_users(connection, lookups):
                if verbosity >= 1:
                    self.stdout.write("Synced {user}".format(
                        user=user,
                    ))



from django.core.management.base import BaseCommand, CommandError
from django.db import transaction
from django.contrib.auth import get_user_model


class Command(BaseCommand):

    help = "Promotes the named users to an admin superuser."

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        parser.add_argument(
            "usernames",
            metavar="usernames",
            nargs="*",
            help="Usernames to promote to admin superuser.",
        )

    @transaction.atomic()
    def handle(self, **kwargs):
        verbosity = kwargs["verbosity"]
        User = get_user_model()
        for username in kwargs["usernames"]:
            try:
                user = User.objects.get(username=username)
            except User.DoesNotExist:
                raise CommandError("User with username {username} does not exist".format(
                    username=username,
                ))
            else:
                user.is_staff = True
                user.is_superuser = True
                user.save()
                if verbosity >= 1:
                    self.stdout.write("Promoted {user} to admin superuser".format(
                        user=user,
                    ))



from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand, CommandError
from django.db import transaction
from django.db.models import ProtectedError

from django_python3_ldap import ldap
from django_python3_ldap.conf import settings
from django_python3_ldap.utils import group_lookup_args


class Command(BaseCommand):

    help = "Remove local user models for users not find anymore in the remote LDAP authentication server."

    def add_arguments(self, parser):
        parser.add_argument(
            '-p',
            '--purge',
            action='store_true',
            help='Purge instead of deactive local user models'
        )
        parser.add_argument(
            'lookups',
            nargs='*',
            type=str,
            help='A list of lookup values, matching the fields specified in LDAP_AUTH_USER_LOOKUP_FIELDS. '
                 'If this is not provided then ALL users are concerned.'
        )
        parser.add_argument(
            '--superuser',
            action='store_true',
            help='Handle superuser (by default, superusers are excluded)'
        )
        parser.add_argument(
            '--staff',
            action='store_true',
            help='Handle staff user (by default,staff users are excluded)'
        )

    @staticmethod
    def _iter_local_users(User, lookups, superuser, staff):
        """
        Iterates over local users. If the list of lookups is empty, then all users are returned.
        However, if lookups are provided, User.object.get is used to clean each user found using the lookups.
        Exclude or not superuser and or staff user.
        """

        if len(lookups) < 1:
            for user in User.objects.filter(is_superuser=superuser,
                                            is_staff=staff):
                yield user
        else:
            for lookup in group_lookup_args(*lookups):
                try:
                    yield User.objects.get(**lookup,
                                           is_superuser=superuser,
                                           is_staff=staff)
                except User.DoesNotExist:
                    raise CommandError("Could not find user with lookup : {lookup}".format(
                        lookup=lookup,
                    ))

    @staticmethod
    def _remove(user, purge):
        """
        Deactivate or purge a given local user
        """
        if purge:
            # Delete local user
            try:
                user.delete()
            except ProtectedError as e:
                raise CommandError("Could not purge user {user} : {e}".format(
                    user=user,
                    e=e
                ))
        else:
            # Deactivate local user
            user.is_active = False
            user.save()

    @transaction.atomic()
    def handle(self, *args, **kwargs):
        verbosity = int(kwargs.get("verbosity", 1))
        purge = kwargs.get('purge', False)
        lookups = kwargs.get('lookups', [])
        superuser = kwargs.get('superuser', False)
        staff = kwargs.get('staff', False)
        User = get_user_model()
        auth_kwargs = {
            User.USERNAME_FIELD: settings.LDAP_AUTH_CONNECTION_USERNAME,
            'password': settings.LDAP_AUTH_CONNECTION_PASSWORD
        }
        with ldap.connection(**auth_kwargs) as connection:
            if connection is None:
                raise CommandError("Could not connect to LDAP server")
            for user in self._iter_local_users(User, lookups, superuser, staff):
                # For each local users
                # Check if user still exists
                user_kwargs = {
                    User.USERNAME_FIELD: getattr(user, User.USERNAME_FIELD)
                }
                if connection.has_user(**user_kwargs):
                    # User still exists on LDAP side
                    continue
                # Clean user
                self._remove(user, purge)
                if verbosity >= 1:
                    self.stdout.write("{action} {user}".format(
                        action=('Purged' if purge else 'Deactivated'),
                        user=user,
                    ))



"""
Some useful LDAP utilities.
"""

import re
import binascii
import itertools

try:
    from django.utils.encoding import force_str
except ImportError:
    from django.utils.encoding import force_text as force_str

from django.utils.module_loading import import_string

from django_python3_ldap.conf import settings


def import_func(func):
    if callable(func):
        return func
    elif isinstance(func, str):
        return import_string(func)
    raise AttributeError("Expected a function {0!r}".format(func))


def clean_ldap_name(name):
    """
    Transforms the given name into a form that
    won't interfere with LDAP queries.
    """
    return re.sub(
        r'[^a-zA-Z0-9 _\-.@:*]',
        lambda c: "\\" + force_str(binascii.hexlify(c.group(0).encode("latin-1", errors="ignore"))).upper(),
        force_str(name),
    )


def convert_model_fields_to_ldap_fields(model_fields):
    """
    Converts a set of model fields into a set of corresponding
    LDAP fields.
    """
    return {
        settings.LDAP_AUTH_USER_FIELDS[field_name]: field_value
        for field_name, field_value
        in model_fields.items()
    }


def format_search_filter(model_fields):
    """
    Creates an LDAP search filter for the given set of model
    fields.
    """
    ldap_fields = convert_model_fields_to_ldap_fields(model_fields)
    ldap_fields["objectClass"] = settings.LDAP_AUTH_OBJECT_CLASS
    search_filters = import_func(settings.LDAP_AUTH_FORMAT_SEARCH_FILTERS)(ldap_fields)
    return "(&{})".format("".join(search_filters))


def clean_user_data(model_fields):
    """
    Transforms the user data loaded from
    LDAP into a form suitable for creating a user.
    """
    return model_fields


def format_username_openldap(model_fields):
    """
    Formats a user identifier into a username suitable for
    binding to an OpenLDAP server.
    """
    return "{user_identifier},{search_base}".format(
        user_identifier=",".join(
            "{attribute_name}={field_value}".format(
                attribute_name=clean_ldap_name(field_name),
                field_value=clean_ldap_name(field_value),
            )
            for field_name, field_value
            in convert_model_fields_to_ldap_fields(model_fields).items()
        ),
        search_base=settings.LDAP_AUTH_SEARCH_BASE,
    )


def format_username_active_directory(model_fields):
    """
    Formats a user identifier into a username suitable for
    binding to an Active Directory server.
    """
    username = model_fields["username"]
    if settings.LDAP_AUTH_ACTIVE_DIRECTORY_DOMAIN:
        username = "{domain}\\{username}".format(
            domain=settings.LDAP_AUTH_ACTIVE_DIRECTORY_DOMAIN,
            username=username,
        )
    return username


def format_username_active_directory_principal(model_fields):
    """
    Formats a user identifier into a username suitable for
    binding to an Active Directory server.
    """
    username = model_fields["username"]
    if settings.LDAP_AUTH_ACTIVE_DIRECTORY_DOMAIN:
        username = "{username}@{domain}".format(
            username=username,
            domain=settings.LDAP_AUTH_ACTIVE_DIRECTORY_DOMAIN,
        )
    return username


def sync_user_relations(user, ldap_attributes, *, connection=None, dn=None):
    # do nothing by default
    pass


def format_search_filters(ldap_fields):
    return [
        "({attribute_name}={field_value})".format(
            attribute_name=clean_ldap_name(field_name),
            field_value=clean_ldap_name(field_value),
        )
        for field_name, field_value
        in ldap_fields.items()
    ]


def group_lookup_args(*args):
    """
    Yields the given series of arguments as chunks, formatted as dictionaries, which represent field lookups
    according to the LDAP_AUTH_USER_LOOKUP_FIELDS setting.

    Based on the itertools grouper recipe: https://docs.python.org/3/library/itertools.html#itertools-recipes
    """
    fields_len = len(settings.LDAP_AUTH_USER_LOOKUP_FIELDS)
    fields = [iter(args)] * fields_len
    for chunk in itertools.zip_longest(*fields, fillvalue=None):
        lookup = {}
        for i in range(fields_len):
            lookup[settings.LDAP_AUTH_USER_LOOKUP_FIELDS[i]] = chunk[i]
        yield lookup
